/* 

		symax0.3.c

    By P.Costigan email: pcscada@vicnet.net.au
    This program contains the protocol for reading and writing
    to a SY/MAX processor (industrial controller) via a serial port.

    Copyright (C) 1998 Philip Costigan   P.C. SCADA LINK PTY. LTD.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



*/

#include <stdio.h>	/* Standard input/output */
#include <sys/types.h>	
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>	/* UNIX standard function definitions */
#include <fcntl.h>	/* File control definitions */
#include <string.h>	/* String function definitions */
#include <errno.h>	/* Error number definitions */
#include <termios.h>	/* POSIX terminal control definitions */

#include "commcodes.h"  /* SY/MAX command codes */
#include "symax0.3.h"   





#ifdef DIAG
/***************************************************************************
***************************************************************************

 The functions PrintRecInfo and PrintSendInfo
 are designed to echo characters, read and writen from
 the Sy/Max device, to the standard error.
 The result is a screen full of characters representing each byte of
 data sent and received. (It takes a bit to decifer).
 This may change down the track to be a seperate diagnostic tool that
 has its own interactive screen. (ie. wish list). 

***************************************************************************
***************************************************************************/


/* Display results to standard error */

void PrintRecInfo( int ReadCode )
{
	fprintf( stderr , "[%0.2X]" , ReadCode );
}

/* Display results to standard error */

void PrintSendInfo( int SendCode )
{
	fprintf( stderr , "<%0.2X>" , SendCode );
}


/***************************************************************************
***************************************************************************/

#endif





#define MAX_FRAME_LENGTH 500
#define NO_DATA -1
#define MAX_ROUTE_STRING 12


/***************************************************************************
****************************************************************************

 For all those people who are going to criticise. I am not a professional C
 programmer. I only get a chance to dabble in this kind of code once or 
 twice a month so forgive me the use of global variables in this code or
 better still, help me tidy this code up.

****************************************************************************
***************************************************************************/

int raw_in[ MAX_FRAME_LENGTH ];		/* raw data received string       */
int raw_in_pos = 0;			/* current position in the string */
int raw_in_ack = 0;			/* acknowledge data in signal     */
int raw_out[ MAX_FRAME_LENGTH ];	/* raw data sent out string       */
int raw_out_pos;			/* current position in the string */
int reply_out[ MAX_FRAME_LENGTH ];	/* reply data sent out string     */
int reply_out_pos;			/* current position in the string */
int reply_out_length;			/* length of reply string         */
int reply_ready = FALSE;		/* Set TRUE when a reply is ready */
int req_out[ MAX_FRAME_LENGTH ];	/* request data from Sy/Max string*/
int req_out_pos;			/* current position in the string */
int req_ready = FALSE;			/* Set TRUE when request is ready */
int ack_flag;				/* set to ODD or EVEN when acking */
int ack_send = FALSE;			/* set TRUE when ready to ack     */
int last_response;			/* remember last reply action     */
int odd_even = ODD;			/* set ODD or EVEN for each d'gram*/
int nak_count = 0;			/* keep count of nak's            */
int syn_count = 0;			/*   "    "   "  syn's            */
int enq_count = 0;			/*   "    "   "  enq's            */
int next_comm_pause = 0;		/* Bad delay tecnique counter     */
int sending = FALSE;			/* TRUE when sending a string     */

void check_raw_in( int *data_string );
int checksum( int *data_string, int bccpos );
void checkroute( int *rawstring );
void checkcommand( int *datastring );


int *trw_status;		/* returned status from SY/MAX              */
int *tread_data_array;		/* Pointer to location to put the data read */
int trw_not_busy = TRUE;	/* status of tread or twrite function       */
int send_hold = FALSE;		/* hold data back while sending ack's       */






/*****************************************************************************
******************************************************************************

 This function merely opens a comm port with default SY/MAX settings.
 1 start bit, 1 stop bit, even parity, 8 data bits.

 I have hard coded this for now and hope to implement a config file sometime
 down the track.

******************************************************************************
*****************************************************************************/

int open_symax_port( char *device )
{
	int fd;
	int p, c;
	struct termios settings;

	int dle = DLE;
	int enq = ENQ;

	if(( fd = open( device, O_RDWR ) ) < 0 )
	{
		fprintf( stderr, "Error opening device %s\n", device );
		exit( 1 );
	}

	tcgetattr( fd, &settings );	/* Get the current setting of device */
	cfsetispeed( &settings, B9600 );/* Set the baud rate */
	cfsetospeed( &settings, B9600 );

	settings.c_cflag |= ( CLOCAL | CREAD );		/* 1 start, 1 stop, */
	settings.c_cflag |= PARENB;		/* even parity, 8 data bits */
	settings.c_cflag &= ~PARODD;
	settings.c_cflag &= ~CSIZE;
	settings.c_cflag |= CS8;
	settings.c_cflag &= ~CRTSCTS;
	settings.c_iflag &= ~ICRNL;
	settings.c_iflag &= ~IXON;
	settings.c_iflag &= ~IGNPAR;
	settings.c_iflag |= PARMRK;
	settings.c_oflag &= ~OPOST;
	settings.c_lflag &= ~( ICANON | ECHO | ISIG );


	tcsetattr( fd, TCSANOW, &settings );
	
	for( p = 0; p < IO_MAP_SIZE; p++ )
		io_map[ p ] = 0;
	
	for( p = 0; p < MAX_FRAME_LENGTH; p++ )
		raw_in[ p ] = NO_DATA;
	raw_in_pos = 0;

	for( c = 0; c < 2; c++ )
	{
		write( fd, &dle, 1 );	/* poll the SY/MAX device */
		write( fd, &enq, 1 );
#ifdef DIAG
		PrintSendInfo( DLE );
		PrintSendInfo( ENQ );
#endif
	}

	last_response = ENQ;
	return( fd );
}




/*****************************************************************************
******************************************************************************
 If you have read this far then you must have an interest in what I have
 done. So I'll try to explain how the protocol and this code works.

 The SY/MAX protocol is based on the ANSI 3.28 standard. You may need to
 search the web to find it. 
 Basically the protocol is designed to be able to send data at
 the same time as data is being received. 
 When a datagram is sent, a reply is expected within a 10 character
 time frame, (even if the replying node is busy sending other stuff).
 
 So my code is designed to take one character at a time and fill up a
 buffer until an end of text signal with checksum arrives. while this
 is happening data can be in the process of being sent (one character 
 at a time). If the checksum of the arrived data is ok, we interupt 
 the sending of data and put an ack in the middle of it. 
 At the end of sending whatever it was that we were sending,
 we then reply to the data that we just acknowledged. (Tricky hey).

 This is just a very basic overview of what is going on. For more info
 you may need to contact Schneider (www.modicon.com) and ask for the
 Square D SY/MAX Instruction Bulletin 30598-351-01

 Its not very accurate so email me if you get confused.

******************************************************************************
*****************************************************************************/


void read_port( int fd )
{
	int inchar = NO_DATA;
	int p;
	fd_set rfds;

	struct timeval tv;

	tv.tv_sec = 0;		/* if no data at serial port then */
	tv.tv_usec = 0;		/* jump straight out of this function */

	FD_ZERO( &rfds );
	FD_SET( fd, &rfds );

	if( select( FD_SETSIZE, &rfds, NULL, NULL, &tv ) ) 
	{
		read( fd, &inchar, 1 );
		inchar  = inchar & 0xFF ;
		raw_in[ raw_in_pos ] = inchar;	/* place data into string */
#ifdef DIAG
		PrintRecInfo( inchar );
#endif

		if( ack_flag && inchar == PAD )
		{
			ack_send = TRUE;
		}

		/* process the raw data string */
		check_raw_in( raw_in );	

		if( raw_in_pos >= MAX_FRAME_LENGTH )
		{
			for( p = 0; p < MAX_FRAME_LENGTH; p++ )
				raw_in[ p ] = NO_DATA;
			raw_in_pos = 0;
		}
		raw_in_pos++;

		next_comm_pause = 0;
	}
	else
	{
		next_comm_pause++;
	}
}







/*****************************************************************************
******************************************************************************

 The function send_port sends characters out of the serial port.

******************************************************************************
*****************************************************************************/

void send_port( int fd )
{
	static int ack_enable = TRUE;
	static int fe_count = 10;

	int dle = DLE;
	int enq = ENQ;
	int n;

	if( ack_send )		/* if safe to send an ack and an ack is */
	{			/* required then send it                */
		if( ack_enable )
		{
			write( fd, &dle, 1 );
			write( fd, &ack_flag, 1 );
#ifdef DIAG
			PrintSendInfo( DLE );
			PrintSendInfo( ack_flag );
#endif
			last_response = ack_flag;
			ack_flag = 0;
			ack_send = FALSE;
			tcdrain( fd );
		}
	}

	if( !sending )		/* If output is idle and a string is to */
	{			/* be sent then ready the string in the */
				/* output buffer.                       */
		if( reply_ready )
		{
			for( n = 0; n < MAX_FRAME_LENGTH; n++ )
			{
				raw_out[ n ] = reply_out[ n ];
			}
			reply_ready = FALSE;
			sending = TRUE;
			raw_in_ack = FALSE;
			raw_out_pos = 0;
		}
		else
 		if( req_ready )
		{
			for( n = 0; n < MAX_FRAME_LENGTH; n++ )
			{
				raw_out[ n ] = req_out[ n ];
			}
			req_ready = FALSE;
			sending = TRUE;
			raw_in_ack = FALSE;
			raw_out_pos = 0;
		}
	}

	if( sending && !send_hold )	/* output data to serial port */ 
	{
		if( raw_out[ raw_out_pos ] == DLE )
		{
			ack_enable = FALSE;
		}
		else
		{
			ack_enable = TRUE;
		}
		write( fd, &raw_out[ raw_out_pos ], 1 );
		tcdrain( fd );
#ifdef DIAG
		PrintSendInfo( raw_out[ raw_out_pos ] );
#endif
		if( raw_out[ raw_out_pos ] == PAD ) 
		{
			fe_count--;
		}
		else
		{
			fe_count = 10;
		}

		if( !fe_count )
		{
			write( fd, &dle, 1 );
			write( fd, &enq, 1 );
#ifdef DIAG
			PrintSendInfo( DLE );
			PrintSendInfo( ENQ );
#endif
			fe_count = 10;
			enq_count++;
			tcdrain( fd );
		}
		raw_out_pos++;
		if( raw_out_pos >= MAX_FRAME_LENGTH )
		{
			sending = FALSE;
			trw_not_busy = TRUE;
		}	

		if( enq_count > 29 )
		{
			sending = FALSE;
			trw_not_busy = TRUE;
			raw_in_ack = FALSE;
			enq_count = 0;
		}

		if( raw_in_ack == raw_out[ 2 ] && 
					raw_out[ raw_out_pos ] == PAD )
		{
			sending = FALSE;
			raw_in_ack = FALSE;
		}
	}
}








/****************************************************************************
*****************************************************************************

 Use this function in a loop to continuously communicate through the 
 serial port.

*****************************************************************************
****************************************************************************/

void symax_comms( int file )
{
	read_port( file );
	send_port( file );
}






/****************************************************************************
*****************************************************************************

 This function processes the raw data that comes in.
 The idea is that as characters come in, this function checks to see if
 anything needs to happen.

*****************************************************************************
****************************************************************************/

void check_raw_in( int *data_string )
{
	int p;
	static int ff_work_around = FALSE;
	static int dle_in_data = FALSE;

	/* prototype */
	void build_reply_header( );

	if( raw_in_pos == 1 )	/* find what datagram ID is being used */
	{
		p = raw_in_pos ;
		if( data_string[ p - 1 ] == DLE )
		{
			if( data_string[ p ] == ODD ||
				data_string[ p ] == EVEN )
			{
				raw_in_ack = data_string[ p ];
			}

			if( data_string[ p ] == NAK )
			{
				
			}
		}
	}



	if( raw_in_pos > 2 )
	{
		p = raw_in_pos ;
		if( data_string[ p - 1 ] == DLE && !dle_in_data )
		{
			switch( data_string[ p ] )
			{
			case SOH:
#ifdef DIAG
				fprintf( stderr, "SOH " );
#endif
				raw_in[ 0 ] = DLE;
				raw_in[ 1 ] = SOH;
				for( p = 2; p < MAX_FRAME_LENGTH; p++ )
				{
					raw_in[ p ] = NO_DATA;
				}
				raw_in_pos = 1;
				p--;
				break;

			case ODD: 
			case EVEN:
#ifdef DIAG
				fprintf( stderr, "ACK ");
#endif
				raw_in_ack = data_string[ p ];
				raw_in_pos -= 2;
				break;

			case NAK:
#ifdef DIAG
				fprintf( stderr, "NAK " );
#endif
				raw_in_pos -= 2;
				raw_out_pos = 0;
				nak_count++;
				break;

			case SYN:
#ifdef DIAG
				fprintf( stderr, "SYN " );
#endif
				raw_in_pos -= 2;
				raw_out_pos = 0;
				syn_count++;
				break;

			case ENQ:
#ifdef DIAG
				fprintf( stderr, "ENQ " );
#endif
				ack_flag = last_response; 	
			 	break;
 			default:
				break;
			}
		}

		p = raw_in_pos ;
		if( data_string[ p - 2 ] == DLE )
		{
			if( data_string[ p - 1 ] == ETX )
			{
#ifdef DIAG
				fprintf( stderr, "ETX " );
#endif
				send_hold = FALSE;
				if( checksum( data_string,raw_in_pos ))
				{
					if( data_string[2] == last_response )
					{
						ack_flag = SYN;
					}
					else
					{
						ack_flag = data_string[ 2 ];
						checkroute( data_string );
					}
					raw_in_pos = 0;
				}
				else
				{
					ack_flag = NAK;
				}
			}
		}

		p = raw_in_pos ;
		if( data_string[ p - 1 ] == DLE )
		{
			if( data_string[ p ] == DLE )
			{
				if( !dle_in_data )
				{
#ifdef DIAG
					fprintf( stderr, "DLE " );
#endif
					dle_in_data = TRUE;
				}
				else
				{
					dle_in_data = FALSE;
				}
			}
			else
			{
				dle_in_data = FALSE;
			}
		}
	
		

		p = raw_in_pos ;
		if( data_string[ p - 1 ] == 0xFF )
		{
			if( data_string[ p ] == 0xFF && !ff_work_around )
			{
#ifdef DIAG
				fprintf( stderr, ".ff." );
#endif
				raw_in_pos--;
				ff_work_around = TRUE;
			}
			else
			{
				ff_work_around = FALSE;
			}
		}
	}
}





/****************************************************************************
*****************************************************************************

 Check the received data for errors.
 The checksum is an eight bit 2's complement of the sum of all the data
 between the SOH and the checksum byte. 

 There is a variation with this formula.
 I'm not sure whether the phenomenon is caused by SY/MAX or by Linux/ttyS.
 If an FF is received then another FF is placed in the string but not included
 in the checksum (just to keep us on our toes). 

*****************************************************************************
****************************************************************************/

int checksum( int *checkstring, int bccpos )
{
	int p;
	int bcc = 0;
	static int last_bcc;
	int ret = FALSE;

	for( p = 2; p < bccpos; p++ )	/* start at the char after SOH */
	{				/* and start adding.	       */
		bcc += checkstring[ p ];
	}

	bcc = ( 0 - bcc ) & 0xFF;	/* Make an 8bit 2's complement */

	if( bcc == last_bcc ) 		/* check if retry atempted     */
	{
		ack_send = SYN;
		send_hold = TRUE;
	}
	else
	{
		send_hold = FALSE;
	}

	last_bcc = bcc;
					/* compare with received checksum */
	if( bcc == checkstring[ bccpos ] )
	{
		ret = TRUE;
	}

	return( ret );
}





/*****************************************************************************
******************************************************************************

 This function is used to toggle the datagram ID

******************************************************************************
*****************************************************************************/

void change_odd_even( void )
{
	if( odd_even == ODD )
	{
		odd_even = EVEN;
	}
	else
	{
		odd_even = ODD;
	}
}




/*****************************************************************************
******************************************************************************

 build_reply_header sets up the first three data bytes in the reply string.

******************************************************************************
*****************************************************************************/

void build_reply_header( void )
{
	change_odd_even();
	reply_out[ 0 ] = DLE;
	reply_out[ 1 ] = SOH;
	reply_out[ 2 ] = odd_even;
}



/*****************************************************************************
******************************************************************************

 build_route takes the network route address received and swaps it around
 for sending the reply back to the right SY/MAX device.

******************************************************************************
*****************************************************************************/

void build_route( int *route_string, int size )
{
	int n;

	reply_out_pos = 3;

	for( n = ( size - 1 ) ; n > 0; n-- )
	{
		reply_out[ reply_out_pos++ ] = route_string[ n - 1 ];
		reply_out[ reply_out_pos++ ] = route_string[ n-- ];
	}
}




/*****************************************************************************
******************************************************************************

 build_write_reply is designed to reply to a SY/MAX device that has written
 data into our main data map.
 The reply here is hard coded to reply with GOOD COMMS. No attempt has been
 made to check for illegal address or receive overflow errors ( YET ).
 So if you conform to the SY/MAX communication rules, things should work out
 fine. But if you go outside the boundaries expect the worst.

******************************************************************************
*****************************************************************************/

void build_write_reply( int *status_address )
{
	void build_checksum( void );

	build_reply_header();

	reply_out[ reply_out_pos++ ] = DLE;
	reply_out[ reply_out_pos++ ] = STX;
	reply_out[ reply_out_pos++ ] = WRITEREPLY;
	reply_out[ reply_out_pos++ ] = status_address[ 0 ];
	reply_out[ reply_out_pos++ ] = status_address[ 1 ];
	reply_out[ reply_out_pos++ ] = 0x80;
	reply_out[ reply_out_pos++ ] = 0x00;
	reply_out[ reply_out_pos++ ] = DLE;
	reply_out[ reply_out_pos++ ] = ETX;

	build_checksum( );

	reply_out_length = reply_out_pos;
	while( reply_out_pos < MAX_FRAME_LENGTH )
	{
		reply_out[ reply_out_pos++ ] = PAD;
	}

}



void build_checksum( void )
{
	int n;
	int bcc = 0;


	for( n = 2; n < reply_out_pos; n++ )
	{
		bcc += reply_out[ n ];
	}


	reply_out[ reply_out_pos++ ] = ( 0 - bcc ) & 0xFF;

	reply_ready = TRUE;
}







/****************************************************************************
*****************************************************************************

 This function checks if route information was received in the datagram.
 If the serial port is connected directly to a processor then no route is
 necessary.

*****************************************************************************
****************************************************************************/

void checkroute( int *rawstring )
{
	int route[ MAX_ROUTE_STRING ] = { 0 };
	int data[ MAX_FRAME_LENGTH ];
	int raw_route_pos = 3;
	int route_pos = 0;
	int data_pos;


	while( rawstring[ raw_route_pos ] != DLE )
	{
		route[ route_pos++ ] = rawstring[ raw_route_pos++ ];
	}
	if( route[ 0 ] ) build_route( route, route_pos );

	data_pos = route_pos + 5;

	checkcommand( (rawstring + data_pos) );
}







/****************************************************************************
*****************************************************************************

 This function performs the command that was received in the datagram.

*****************************************************************************
****************************************************************************/

void checkcommand( int *datastring )
{
	void writecommand( int *datastring );
	void readcommand( int *datastring );
	void readreply( int *datastring );
	void writereply( int *datastring );

	switch( datastring[ 0 ] )
	{
		case WRITECOMM:
			writecommand( datastring + 1 );
			break;
		case READCOMM:
			readcommand( datastring + 1 );
			break;
		case READREPLY:
			readreply( datastring + 1 );
			break;
		case WRITEREPLY:
			writereply( datastring + 1 );
			break;
	}
	reply_out_pos = 0;

	enq_count = 0;

}






/*****************************************************************************
******************************************************************************

 writecommand takes in the data received and sorts out where the data is to
 be placed in the main data map.

******************************************************************************
*****************************************************************************/

void writecommand( int *datastring )
{
	int statusreg[2];
	int startaddress;

	void writedata( int location, int *datastring );
	void build_write_reply( int *statusreg );

	sending = FALSE;

	statusreg[0] = datastring[0];
	statusreg[1] = datastring[1];

	startaddress = ((((datastring[2] << 8 ) + datastring[3] ) + 2 ) / 2 );

	writedata( startaddress, datastring + 4 );
	build_write_reply( statusreg );
}






/*****************************************************************************
******************************************************************************

 writedata places the data receive into the main data map. It also filters
 out ack's nak's and syn's etc.

******************************************************************************
*****************************************************************************/

void writedata( int location, int *datastring )
{
	int datastringloc = 0;
	int finished = FALSE;
	int msb = TRUE;

	while( !finished )
	{
		if( datastring[ datastringloc ] == DLE )
		{
			switch( datastring[ ++datastringloc ] )
			{
				case DLE:
					break;
				case ODD:
				case EVEN:
				case NAK:
				case SYN:
					datastringloc++;
					break;
				case ETX:
					finished = TRUE;
					break;
				default:
					break;
			}
		}
		if( !finished )
		{
			if( msb )
			{
				io_map[ location ] = (short)
					( datastring[ datastringloc ] << 8 );
				msb = FALSE;
			}
			else
			{
				io_map[ location ] |= (short)
						datastring[ datastringloc ]; 
				location++;
				msb = TRUE;
			}
			datastringloc++;
		}
		
	}
	
}





/*****************************************************************************
******************************************************************************

 I'm not sure whether to implement this or not yet. If implemented it would
 mean that other SY/MAX devices could request data from the main data map.

******************************************************************************
*****************************************************************************/

void readcommand( int *datastring )
{

	fprintf( stderr, " Read command " );
	fprintf( stderr, " NOT IMPLEMENTED YET " );
}





/*****************************************************************************
******************************************************************************

 readreply is the command received when this code requests data from another
 SY/MAX device. This function takes the data and puts it into the location 
 that has been pointed to by the tread function.

******************************************************************************
*****************************************************************************/

void readreply( int *datastring )
{

	void readreplydata( int *datastring );

	*trw_status = datastring[ 2 ] << 8;
	*trw_status |= datastring[ 3 ];

	readreplydata( datastring + 6 );
}





/*****************************************************************************
******************************************************************************

 Effectivly this function does exactly the same as the function writedata.
 The difference is that writedata only puts data into the main data map and
 this function can put data into private data maps.
 Maybe this needs a bit more thought because I believe these two could be one
 function. I'm just being slack for now.

******************************************************************************
*****************************************************************************/

void readreplydata( int *datastring )
{
	int datastring_loc = 0;
	int user_array_loc = 0;
	int finished = FALSE;
	int msb = TRUE;
	
	while( !finished )
	{
		if( datastring[ datastring_loc ] == DLE )
		{
			switch( datastring[ ++datastring_loc ] )
			{
				case DLE:
					break;
				case ODD:
				case EVEN:
				case NAK:
				case SYN:
					datastring_loc++;
					break;
				case ETX:
					finished = TRUE;
					break;
				default:
					break;	

			}
		}
		if( !finished )
		{
			if( msb )
			{
				tread_data_array[ user_array_loc ] = ( short )
					( datastring[ datastring_loc ] << 8 );
				msb = FALSE;
			}
			else
			{
				tread_data_array[ user_array_loc ] |= ( short )
					datastring[ datastring_loc ];
				user_array_loc++;
				msb = TRUE;
			}
			datastring_loc++;
		}

	}
	trw_not_busy = TRUE; 
}





/*****************************************************************************
******************************************************************************

 When this code writes data to A SY/MAX device, the reply is just a good or
 bad signal. The only data to contend with is the status register data.

******************************************************************************
*****************************************************************************/

void writereply( int *datastring )
{

	*trw_status = datastring[ 2 ] << 8;
	*trw_status |= datastring[ 3 ];
	
	trw_not_busy = TRUE;

}






/*****************************************************************************
******************************************************************************

 tread is the function to use to request data from a SY/MAX device.
 I have deliberatly tried to keep it to the same style as the SY/MAX ladder
 style so as to keep some similarity between the two systems.

******************************************************************************
*****************************************************************************/

int tread( int *status, int *route, int *local, int remote, int count )
{
	int ret = FALSE;

	/* declaration */
	void build_tread( int *status, int *route, int *local, 
					int remote, int count  );

	if( trw_not_busy && !send_hold && next_comm_pause > 150 )
	{
		count--;
		build_tread( status, route, local, remote, count ); 
		ret = TRUE;
		trw_not_busy = FALSE;
		next_comm_pause = 0;
	}

	if( next_comm_pause > 190 )
	{
		trw_not_busy = TRUE;
		send_hold = FALSE;
	}
	return( ret );
}





/*****************************************************************************
******************************************************************************

 twrite is the function to use when sending data to a SY/MAX device.
 I have deliberatly tried to keep it to the same style as the SY/MAX ladder
 style so as to keep some similarity between the two systems. (As above).

******************************************************************************
*****************************************************************************/

int twrite( int *status, int *route, int *local, int remote, int count )
{
	int ret = FALSE;

	void build_twrite( int *status, int *route, int *local,
					int remote, int count  );

	if( trw_not_busy && !send_hold && next_comm_pause > 150 )
	{
		build_twrite( status, route, local, remote, count );
		ret = TRUE;
		trw_not_busy = FALSE;
		next_comm_pause = 0;
	}

	if( next_comm_pause > 190 )
	{
		trw_not_busy = TRUE;
		send_hold = FALSE;
	}
	return( ret );
}





/*****************************************************************************
******************************************************************************

 build_trw_header sets up the first three data bytes in the string to send.

******************************************************************************
*****************************************************************************/

void build_trw_header( void )
{
	change_odd_even();
	req_out[ req_out_pos++ ] = DLE;
	req_out[ req_out_pos++ ] = SOH;
	req_out[ req_out_pos++ ] = odd_even;

}




/*****************************************************************************
******************************************************************************

 build_trw_route builds the string needed to network the data to the correct
 SY/MAX device. Used with tread and twrite.

******************************************************************************
*****************************************************************************/

void build_trw_route( int *route_string )
{
	int n = 0;
	int ascii_1;
	int ascii_2;

	if( route_string )
	{	
		while( route_string[ n ] )
		{
			/* convert route into ascii equivelent */
			ascii_1 = route_string[ n ] >> 4;
			if( ascii_1 < 0x0A ) 
				ascii_1 += 0x30;
			else
				ascii_1 += 0x37;

			ascii_2 = route_string[ n ] & 0x0F;
			if( ascii_2 < 0x0A )
				ascii_2 += 0x30;
			else
				ascii_2 += 0x37;
		
			req_out[ req_out_pos++ ] = ascii_1;
			req_out[ req_out_pos++ ] = ascii_2;
			n++;
		}
	}
}





void build_trw_checksum( void )
{
	int n;
	int bcc = 0;

	for( n = 2; n < req_out_pos; n++ )
	{
		bcc += req_out[ n ];
	}

	req_out[ req_out_pos++ ] = ( 0 - bcc ) & 0xFF;

	req_ready = TRUE;	
}





/*****************************************************************************
******************************************************************************

 This function checks for the value 10 hex which is a DLE and appends another
 10 hex after it in the string. this is necessary to inform the receiver that
 the 10 hex is infact the value 10 hex and not a command ( DLE ).

******************************************************************************
*****************************************************************************/

void build_trw_checkforDLE( void )
{
	req_out_pos--;
	if( req_out[ req_out_pos ] == DLE )
	{
		req_out_pos++;
		req_out[ req_out_pos ] = DLE;
	}

	req_out_pos++;
}





/*****************************************************************************
******************************************************************************

 This function puts together the string required to point to the register
 group to read from and how many to get.

******************************************************************************
*****************************************************************************/

void build_tread_req( int *local_string, int remote_reg, int count )
{
	int temp;

	req_out[ req_out_pos++ ] = DLE;
	req_out[ req_out_pos++ ] = STX;
	req_out[ req_out_pos++ ] = READCOMM;
	req_out[ req_out_pos++ ] = 0x00;
	req_out[ req_out_pos++ ] = 0x00;
	req_out[ req_out_pos++ ] = 0x00;
	req_out[ req_out_pos++ ] = 0x02;

	temp = ( remote_reg * 2 ) - 2;
	req_out[ req_out_pos++ ] = temp >> 8;
	build_trw_checkforDLE( );
	req_out[ req_out_pos++ ] = temp & 0xFF;
	build_trw_checkforDLE( );	
	req_out[ req_out_pos++ ] = count >> 8;
	build_trw_checkforDLE( );
	req_out[ req_out_pos++ ] = count & 0xFF;	
	build_trw_checkforDLE( );
	req_out[ req_out_pos++ ] = DLE;
	req_out[ req_out_pos++ ] = ETX;

	tread_data_array = local_string;

	build_trw_checksum( );

	while( req_out_pos < MAX_FRAME_LENGTH )
	{
		req_out[ req_out_pos++ ] = PAD;
	}

	req_out_pos = 0;
}






/*****************************************************************************
******************************************************************************

 This function compiles all the data to send and which register group to put
 it and places it into the string to send.

******************************************************************************
*****************************************************************************/

void build_twrite_req( int *local_string, int remote_reg, int count )
{
	int temp, n;

	req_out[ req_out_pos++ ] = DLE;
	req_out[ req_out_pos++ ] = STX;
	req_out[ req_out_pos++ ] = WRITECOMM;
	req_out[ req_out_pos++ ] = 0x00;
	req_out[ req_out_pos++ ] = 0x00;

	temp = ( remote_reg * 2 ) - 2;
	req_out[ req_out_pos++ ] = temp >> 8;
	build_trw_checkforDLE( );
	req_out[ req_out_pos++ ] = temp & 0xFF;
	build_trw_checkforDLE( );

	for( n = 0; n < count; n++ )
	{
		req_out[ req_out_pos++ ] = local_string[ n ] >> 8 & 0xFF;
		build_trw_checkforDLE( );
		req_out[ req_out_pos++ ] = local_string[ n ] & 0xFF;
		build_trw_checkforDLE( );
	} 

	req_out[ req_out_pos++ ] = DLE;
	req_out[ req_out_pos++ ] = ETX;

	build_trw_checksum( );

	while( req_out_pos < MAX_FRAME_LENGTH )
	{
		req_out[ req_out_pos++ ] = PAD;
	}
	req_out_pos = 0;
}





void build_tread( int *status, int *route, int *local, int remote, int count )
{
	trw_status = status;

	build_trw_header( );
	build_trw_route( route );
	build_tread_req( local, remote, count );

}


void build_twrite( int *status, int *route, int *local, int remote, int count )
{
	trw_status = status;

	build_trw_header( );
	build_trw_route( route );
	build_twrite_req( local, remote, count );
}








